unit ICComponentResources;

(* Start Header *)

(* File:         ComponentResources.p
 * Generated by: 1.0d6
 * For:          IC 2.0
 * On:           Sunday, 26 October 1997, 17:20:55
 * 
 * This file is part of the Internet Configuration system and
 * is placed in the public domain for the benefit of all.
 *)

(* End Header *)

(* This unit holds the selectors for the IC component.  I've moved this into a unit so
	that the override components can access this information without having to
	duplicate it.
*)

interface

	uses
		Types,
		
		InternetConfig,
		
		ICSystemGlobals;
		
	function SelectorToStr (what: integer): Str32;
	(* This routine converts a IC component into its corresponding identifier
		string.  It is great for debugging because you don't have to remember
		what all of those random selectors are.
	*)

	(* ***** Resources and Resource Management ***** *)

	// Internet Config preferences are stored as resources of
	// type 'ICxy', where x is a character in the range
	// 'R'..'Z' and y is a character in the range 'P'..'Z'.
	// Each combination represents a different "profile".
	// This results in up to 99 profiles, which is a good
	// number because resource files are limited to 2727
	// resources, and many more profiles would put us in
	// serious danger of hitting that limit on a regular
	// basis.
	
	const
		kICPrefResourceTypeBase = 'ICRP';
		kNumberOfDigitsInThirdChar = 9;		(* 'R'..'Z' *)
		kNumberOfDigitsInFourthChar = 11;	(* 'P'..'Z' *)
		
	const
		(* STR# *)
		rMiscStrings = 1024;
			strDamaged = 1;
			strDamagedBeyondRepair = 2;
			strBackupInTrash = 3;
			strDamagedSuffix = 4;
			strDefaultFileName = 5;
			strNewConfigPrompt = 6;
			strCopy = 7;				// " copy"
			strCopyI = 8;				// " copy "
			strDefaultSetName = 9;		// "My Settings"
			
			// If you alter strDefaultSetName you should also
			// alter the 'Pref' resource with key kICProfileName
			// because that contains the default name used for
			// profiles that are converted from old IC pre-2.0
			// preference files.

		(* ALRT *)
		rDamagedAlert = 1024;
		
		(* PROC *)
		rGestaltProc = 129;
			kGestaltProcDataOffset = 2;		// storage for value is 2 bytes into the resource
			
		(* DATA *)
		rICApplicationCreator = 128;
		
		(* CURP *)
		rCurrentProfileResource = 128;

		kICProfileName = '49434170ProfileName';
		kICProfileVersion = '49434170ProfileVersion';
	
{$options align=mac68k}

	type
		ICProfileVersionResource =
			record
				attr : ICAttr;
				version : NumVersion;
			end;
		ICProfileVersionResourcePtr = ^ICProfileVersionResource;
		ICProfileVersionResourceHandle = ^ICProfileVersionResourcePtr;

{$options align=reset}
		
	(* The component regularly needs access to its own resources.
		In some cases, subroutines need access to resources but they don't
		know whether the calling routine has already opened the resource
		file.  [The Component Manager's OpenComponentResFile routine
		makes no attempt to reference count the resource file.]  So we
		build our own reference counting mechanims.
		
		You can use the following two routines in two different ways:
		
			err := StartUsingComponentResources(didOpen);
			if err = noErr begin
				// do stuff
				StopUsingComponentResources(didOpen)
			end
		
		or
		
			err := StartUsingComponentResources(didOpen);
			if err = noErr begin
				// do stuff
			end
			StopUsingComponentResources(didOpen)
		
		Both of these should work just hunky dory.
	*)

	function StartUsingComponentResources(globals: ComponentGlobalsPtr; var didOpen : Boolean) : ICError;
		(* Makes the component's resources available.  Sets didOpen to true if it actually
			opened the resource file.  If this returns noErr, you should make sure you make
			a matching call to StopUsingComponentResources.
		*)

	function StopUsingComponentResources(globals: ComponentGlobalsPtr; didOpen : Boolean) : ICError;
		(* Closes the component's resource file if we opened it. *)

	function GetComponentIndString (globals: ComponentGlobalsPtr; 
									strListID: integer; index: integer; str: StringPtr): ICError;
		(* Get a string from a STR# resource out of the component resource fork.
			One important feature of this routine is that it doesn't modify str unless
			it succeeds in getting a non-empty string.  This is to assist clients
			that go:
				str := 'hardwire constant';
				err := GetComponentIndString(...);
			and junk the error code, assuming that the hardwired constant is
			good enough in the presence of errors.
		*)
	
implementation

	uses
		Components,
		Resources,
		TextUtils,
		
		InternetConfig;

	function SelectorToStr (what: integer): Str32;
		(* See comment in interface part. *)
	begin
		case what of
			(* Component Manager stuff *)
			kComponentVersionSelect: 
				SelectorToStr := 'kComponentVersionSelect';
			kComponentCanDoSelect: 
				SelectorToStr := 'kComponentCanDoSelect';
			kComponentOpenSelect: 
				SelectorToStr := 'kComponentOpenSelect';
			kComponentCloseSelect: 
				SelectorToStr := 'kComponentCloseSelect';
			kComponentTargetSelect: 
				SelectorToStr := 'kComponentTargetSelect';
			kComponentRegisterSelect: 
				SelectorToStr := 'kComponentRegisterSelect';
			kComponentUnregisterSelect: 
				SelectorToStr := 'kComponentUnregisterSelect';
			(* this component type stuff *)
(* Start ICCSelToStr.p *)

			kICCStart: 
			    SelectorToStr := 'kICCStart';
			kICCStop: 
			    SelectorToStr := 'kICCStop';
			kICCFindConfigFile: 
			    SelectorToStr := 'kICCFindConfigFile';
			kICCFindUserConfigFile: 
			    SelectorToStr := 'kICCFindUserConfigFile';
			kICCGeneralFindConfigFile: 
			    SelectorToStr := 'kICCGeneralFindConfigFile';
			kICCChooseConfig: 
			    SelectorToStr := 'kICCChooseConfig';
			kICCChooseNewConfig: 
			    SelectorToStr := 'kICCChooseNewConfig';
			kICCGetConfigName: 
			    SelectorToStr := 'kICCGetConfigName';
			kICCGetConfigReference: 
			    SelectorToStr := 'kICCGetConfigReference';
			kICCSetConfigReference: 
			    SelectorToStr := 'kICCSetConfigReference';
			kICCSpecifyConfigFile: 
			    SelectorToStr := 'kICCSpecifyConfigFile';
			kICCRefreshCaches: 
			    SelectorToStr := 'kICCRefreshCaches';
			kICCGetSeed: 
			    SelectorToStr := 'kICCGetSeed';
			kICCGetPerm: 
			    SelectorToStr := 'kICCGetPerm';
			kICCDefaultFileName: 
			    SelectorToStr := 'kICCDefaultFileName';
			kICCBegin: 
			    SelectorToStr := 'kICCBegin';
			kICCGetPref: 
			    SelectorToStr := 'kICCGetPref';
			kICCSetPref: 
			    SelectorToStr := 'kICCSetPref';
			kICCFindPrefHandle: 
			    SelectorToStr := 'kICCFindPrefHandle';
			kICCGetPrefHandle: 
			    SelectorToStr := 'kICCGetPrefHandle';
			kICCSetPrefHandle: 
			    SelectorToStr := 'kICCSetPrefHandle';
			kICCCountPref: 
			    SelectorToStr := 'kICCCountPref';
			kICCGetIndPref: 
			    SelectorToStr := 'kICCGetIndPref';
			kICCDeletePref: 
			    SelectorToStr := 'kICCDeletePref';
			kICCEnd: 
			    SelectorToStr := 'kICCEnd';
			kICCEditPreferences: 
			    SelectorToStr := 'kICCEditPreferences';
			kICCParseURL: 
			    SelectorToStr := 'kICCParseURL';
			kICCLaunchURL: 
			    SelectorToStr := 'kICCLaunchURL';
			kICCMapFilename: 
			    SelectorToStr := 'kICCMapFilename';
			kICCMapTypeCreator: 
			    SelectorToStr := 'kICCMapTypeCreator';
			kICCMapEntriesFilename: 
			    SelectorToStr := 'kICCMapEntriesFilename';
			kICCMapEntriesTypeCreator: 
			    SelectorToStr := 'kICCMapEntriesTypeCreator';
			kICCCountMapEntries: 
			    SelectorToStr := 'kICCCountMapEntries';
			kICCGetIndMapEntry: 
			    SelectorToStr := 'kICCGetIndMapEntry';
			kICCGetMapEntry: 
			    SelectorToStr := 'kICCGetMapEntry';
			kICCSetMapEntry: 
			    SelectorToStr := 'kICCSetMapEntry';
			kICCDeleteMapEntry: 
			    SelectorToStr := 'kICCDeleteMapEntry';
			kICCAddMapEntry: 
			    SelectorToStr := 'kICCAddMapEntry';
			kICCGetCurrentProfile: 
			    SelectorToStr := 'kICCGetCurrentProfile';
			kICCSetCurrentProfile: 
			    SelectorToStr := 'kICCSetCurrentProfile';
			kICCCountProfiles: 
			    SelectorToStr := 'kICCCountProfiles';
			kICCGetIndProfile: 
			    SelectorToStr := 'kICCGetIndProfile';
			kICCGetProfileName: 
			    SelectorToStr := 'kICCGetProfileName';
			kICCSetProfileName: 
			    SelectorToStr := 'kICCSetProfileName';
			kICCAddProfile: 
			    SelectorToStr := 'kICCAddProfile';
			kICCDeleteProfile: 
			    SelectorToStr := 'kICCDeleteProfile';
			kICCRequiresInterruptSafe: 
			    SelectorToStr := 'kICCRequiresInterruptSafe';
			kICCGetMapEntryInterruptSafe: 
			    SelectorToStr := 'kICCGetMapEntryInterruptSafe';

(* End ICCSelToStr.p *)
			otherwise
				SelectorToStr := 'other';
		end; (* case *)
	end; (* SelectorToStr *)

	(* ***** Resources and Resource Management ***** *)

	function StartUsingComponentResources(globals: ComponentGlobalsPtr; var didOpen : Boolean) : ICError;
		(* See comment in interface part. *)
		var
			err : ICError;
	begin
		err := noErr;
		if globals^.resourceFileRefNum <> 0 then begin
			didOpen := false;
		end else begin
			globals^.oldResourceFileRefNum := CurResFile;
			globals^.resourceFileRefNum := OpenComponentResFile(Component(globals^.self));
			if globals^.resourceFileRefNum <= 0 then begin
				globals^.resourceFileRefNum := 0;
				err := resNotFound;
			end; (* if *)
			didOpen := (err = noErr);
		end; (* if *)
		StartUsingComponentResources := err;
	end; (* StartUsingComponentResources *)

	function StopUsingComponentResources(globals: ComponentGlobalsPtr; didOpen : Boolean) : ICError;
		(* See comment in interface part. *)
		var
			err : ICError;
	begin
		err := noErr;
		if didOpen then begin
			err := CloseComponentResFile(globals^.resourceFileRefNum);
			if err = noErr then begin
				globals^.resourceFileRefNum := 0;
				UseResFile(globals^.oldResourceFileRefNum);
			end; (* if *)
		end; (* if *)
		StopUsingComponentResources := err;
	end; (* StopUsingComponentResources *)

	function GetComponentIndString (globals: ComponentGlobalsPtr; 
									strListID: integer; index: integer; str: StringPtr): ICError;
		(* See comment in interface part. *)
		var
			err: ICError;
			junk: ICError;
			didOpen : Boolean;
			tmpStr : Str255;
	begin
		err := StartUsingComponentResources(globals, didOpen);
		if err = noErr then begin
			GetIndString(tmpStr, strListID, index);
			if tmpStr <> '' then begin
				(* Only copy out the required number of bytes, ie the number
					of bytes that are actually in the string. This is because
					lots of our clients pass in things that aren't Str255's
					(ie Str31s and Str63s), and we don't want to blat
					beyond what they've allocated.
				*)
				BlockMoveData(@tmpStr[0], str, length(tmpStr) + 1);
			end; (* if *)
			junk := StopUsingComponentResources(globals, didOpen);
		end; (* if *)
		GetComponentIndString := err;
	end; (* GetComponentIndString *)

end. (* ICComponentResources *)
